/**************************************************************

File Name:	CASCADE_MC.asm
		
Date Modified:	05/22/00		GJO
				05/25/00		GJO	Code verified on rev 0.1 Silicon
				09/19/04		JDT Ported to VDSP++ 3.5. replace .SECTIONsyntax

Description:
			ADSP-21364 multichannel SIMD subroutine implements a biquad, 
			canonical form, IIR filter.  The coeficient, input and output 
			buffers are	interleaved.

			Equation:
			w(n) = x(n) + a1*w(n-1) + a2*w(n-2)
			y(n) = w(n) + b1*w(n-1) + b2*w(n-2)	(single biquad structure)

Calling Parameters:
			r0 = 2*SECTIONS
			r1 = SAMPLES
			r5 = 4*SECTIONS
			b0 = pointer to delay line array
		    b1 = pointer to delay line array
			b3 = pointer to input array
			b4 = pointer to output array
			b8 = pointer to coeficient array 
			m1 = 2 modify for delay line
			m2 = 1 modify for IIR init
			m8 = 2 modify for coeficient
		    l0 = 2*SECTIONS*2
			l1, l3, l4 = 0
			l8 = 2*SECTIONS*4
		    

Assumptions:
		 	The coefficient buffer should be in the order a2,a2,a1,a1,b2,b2,b1,b1 for each
			biquad section.  The delay line should be ordered w(n-2), w(n-1) for 
			each biquad section.  Interleaved array of coeficients is applied to each
		 	input sample.
			Delay line, Output and Input buffers stored in Block 1.
			Instructions, and Coefficients stored in Block 0.

			Block 0:
			Coefficient buffer = { a2[0],a2[0],a1[0],a1[0],b2[0],b2[0],b1[0],b1[0],... }
		
			Block 1:
			Input buffer ={ A[0],B[0],A[1],B[1],... }
			Output buffer ={ A'[0],B'[0],A'[1],B'[1],... }


Return Values:
		    f8 = output sample y(n)

Registers Affected:
		    f2, f3, f4, f8, f12
			i0, i1, i3, i4, i8

Cycle Count:
		    5 + SAMPLES(8 + (4*SECTIONS)) + 6 cached        IIR
		    3 + 4*SECTIONS			       					IIR_INIT
			
Memory Usage:
			Instructions Words (48-bits):
				17(iir) + 9(fir_init) instruction words
			
			Data Words (16, 32, 40, or 64-bits):
				2*coefs = Number of locations for coefficients (32-bits)
				4*SECTIONS = Number of locations for the delay line (32-bits)
				2*SAMPLES = Number of samples locations for the input (32-bits)
				2*SAMPLES = Number of samples locations for the output (32-bits)
				

Notes: 
			Circular Buffering Issue:
			Because SIMD or Long word access transfer two 32-bit words, programs
			must be careful when using these accesses in circular buffering.
			It is important that SIMD or Long word accesses do not cross a circular
		 	buffer boundary.  If a SIMD	mode access occurs using a circular buffer
			index register that points to the last location in the circular buffer
		 	(end of buffer), the resulting access transfers the last location in
		 	the circular buffer and the first location outside the buffer (end of 
			buffer + 1).

****************************************************************************/
#include "def21364.h"

.GLOBAL cascaded_biquad, cascaded_biquad_init;
.EXTERN coefs, dline;

.SECTION/PM    seg_pmco;

cascaded_biquad:	bit set mode1 CBUFEN | PEYEN;				/* Enable SIMD mode */
    lcntr=r1, do filtering until lce;
	b0 = dline;
	b1= dline;
    b8 = coefs;
	f8=dm(i3,m1);											/* load input sample */
	r12=r12 xor r12,			f2=dm(i0,m1), 	f4=pm(i8,m8);   /* r12=0, get w(n-2) and a2 */
   	lcntr=r0, do quads until lce;                           /* do: for each section */
    f12=f2*f4, 		f8=f8+f12, 	f3=dm(i0,m1), 	f4=pm(i8,m8);   /* w(n-2)xa2, get w(n-1), get a1 */
    f12=f3*f4, 		f8=f8+f12, 	dm(i1,m1)=f3, 	f4=pm(i8,m8);   /* w(n-1)xa1, x(n)+(w(n-2)xa2), dline=w(n-1), get b2 */ 
    f12=f2*f4, 		f8=f8+f12, 	f2=dm(i0,m1), 	f4=pm(i8,m8);   /* w(n-2)xb2, w(n) = x(n)+w(n-1)xa1+w(n-2)xa2, get next w(n-2), get b1 */
quads:  f12=f3*f4, 	f8=f8+f12, 	dm(i1,m1)=f8, 	f4=pm(i8,m8);   /* w(n-1)xb1, w(n)+(w(n-2)xb2), dline=w(n), get a2 */
        			f8=f8+f12;                               /* calc last y after dropping out of loop */
filtering:  					dm(i4,m1)=f8;
    rts (db);
	bit clr mode1 CBUFEN | PEYEN;				/* disable SIMD mode */
	nop;

cascaded_biquad_init:   /* Initialize dline */
    r3=r3 xor r3;
    lcntr=r5, do clear until lce;                       	/*for each section, do:*/
clear:  dm(i0,m2)=r3;                                    	/* clear w(n-1) and w(n-2) */
    rts (db);
	nop;
	nop;
    

